from typing import List
from datasets import load_dataset

# TODO: to be deleted
from transformers import AutoModelForCausalLM, AutoTokenizer

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import DeepEvalBaseBenchmark
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.big_bench_hard.task import BigBenchHardTask
from deepeval.benchmarks.big_bench_hard.template import BigBenchHardTemplate
from deepeval.scorer import Scorer


class BigBenchHard(DeepEvalBaseBenchmark):
    def __init__(self, tasks: List[BigBenchHardTask] = None):
        super().__init__()
        self.tasks: List[BigBenchHardTask] = tasks
        self.scorer = Scorer()

    def evaluate(self, model: DeepEvalBaseLLM):
        for task in self.tasks:
            goldens = self.load_benchmark_dataset(task)
            total_predictions = len(goldens)
            correct_predictions = 0
            for golden in goldens:
                if self.predict(model, golden):
                    correct_predictions += 1

            print(
                f"Result for Big Bench Hard (task={task.value}: {correct_predictions/total_predictions}"
            )

    def predict(
        self, model: DeepEvalBaseLLM, task: BigBenchHardTask, golden: Golden
    ):
        ##### Example of using the BOOLEAN_EXPRESSIONS task #####

        ##### 1. Define prompt template IF NECESSARY to confine output format ######
        ##### (Only define your own if not found in the original papers) #####
        prompt: dict = BigBenchHardTemplate.generate_output(
            input=input, task=task
        )
        prediction = model(prompt)

        ##### 2. Define metrics IF NECESSARY to evaluate prediction #####
        ##### (Only define metrics if not found in the origianl papers) #######
        return self.scorer.exact_match_score(golden.expected_output, prediction)

    def load_benchmark_dataset(self, task: BigBenchHardTask) -> List[Golden]:
        # load from hugging face
        dataset = load_dataset("lukaemon/bbh", task.value)
        goldens: List[Golden] = []
        for data in dataset["test"]:
            golden = Golden(input=data["input"], expectedOutput=data["target"])
            goldens.append(golden)

        return goldens


###################################
### Example Usage, delete later ###
###################################
class Mistral7B(DeepEvalBaseLLM):
    def __init__(self, model):
        self.model = model
        self.tokenizer = AutoTokenizer.from_pretrained(
            "mistralai/Mistral-7B-v0.1"
        )

    def load_model(self):
        return self.model

    def _call(self, prompt: str) -> str:
        model = self.load_model()

        device = "cuda"  # the device to load the model onto

        model_inputs = self.tokenizer([prompt], return_tensors="pt").to(device)
        model.to(device)

        generated_ids = model.generate(
            **model_inputs, max_new_tokens=100, do_sample=True
        )
        return self.tokenizer.batch_decode(generated_ids)[0]

    def get_model_name(self):
        return "Mistral 7B"


model = AutoModelForCausalLM.from_pretrained("mistralai/Mistral-7B-v0.1")


#############################
######## Example Usage ######
#############################

######## 1. Create Custom Model to be evaluated #################
mistral_7b = Mistral7B(model=model)

######## 2. Define benchmark with tasks #############
benchmark = BigBenchHard(tasks=[BigBenchHardTask.BOOLEAN_EXPRESSIONS])

######## 3. Provide custom model to be evaluated ##########
benchmark.evaluate(model=mistral_7b)

### Done! ###



from typing import List, Optional, Union
import os
import csv
import json
from threading import Lock
from pydantic import BaseModel
import datetime
from concurrent.futures import ThreadPoolExecutor, as_completed

from deepeval.synthesizer.template import EvolutionTemplate, SynthesizerTemplate
from deepeval.models import (
    GPTModel,
    DeepEvalBaseLLM,
    DeepEvalBaseEmbeddingModel,
)
from deepeval.progress_context import synthesizer_progress_context
from deepeval.utils import trimAndLoadJson
from deepeval.dataset.golden import Golden


valid_file_types = ["csv", "json"]


class SyntheticData(BaseModel):
    input: str
    expected_output: str


class Synthesizer:
    def __init__(
        self,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        # embedder: Optional[Union[str, DeepEvalBaseEmbeddingModel]] = None,
        multithreading: bool = True,
        # batch_size: int = 50,
    ):
        if isinstance(model, DeepEvalBaseLLM):
            self.model = model
        else:
            self.model = GPTModel(model=model)

        # self.embedder = embedder
        self.generator_model = self.model.get_model_name()
        self.multithreading = multithreading
        # self.batch_size = batch_size
        self.synthetic_goldens: List[Golden] = []

    def _generate(
        self,
        context: List[str],
        goldens: List[Golden],
        max_goldens_per_context: int,
        lock: Lock,
    ):
        prompt = SynthesizerTemplate.generate_synthetic_data(
            context=context, max_goldens_per_context=max_goldens_per_context
        )
        res = self.model(prompt)
        data = trimAndLoadJson(res)
        synthetic_data = [SyntheticData(**item) for item in data["data"]]
        temp_goldens: List[Golden] = []
        for data in synthetic_data:
            # TODO: evolution

            golden = Golden(
                input=data.input,
                expectedOutput=data.expected_output,
                context=context,
            )
            temp_goldens.append(golden)

        with lock:
            goldens.extend(temp_goldens)

    # TODO
    def generate_goldens(
        self, contexts: List[List[str]], max_goldens_per_context: int = 2
    ) -> List[Golden]:
        with synthesizer_progress_context(self.generator_model):
            goldens: List[Golden] = []

            # 1. get embeddings for context eg., [1,2,3,4,5]
            # 2. group randomly based on embedding similarity eg., [[1,2], [5,2], [4], [3,1,5]]
            # 3. supply as context, generate for each List[str]
            # 4. generation can happen in batches, and each batch can be processed in threads
            # 5. optional evolution
            # 6. optional review
            # 7. return goldens

            # TODO: logic to group and vary contexts

            if self.multithreading:
                lock = Lock()

                with ThreadPoolExecutor() as executor:
                    futures = {
                        executor.submit(
                            self._generate,
                            context,
                            goldens,
                            max_goldens_per_context,
                            lock,
                        ): context
                        for context in contexts
                    }

                    for future in as_completed(futures):
                        future.result()
            else:
                for context in contexts:
                    prompt = SynthesizerTemplate.generate_synthetic_data(
                        context=context,
                        max_goldens_per_context=max_goldens_per_context,
                    )
                    res = self.model(prompt)
                    data = trimAndLoadJson(res)
                    synthetic_data = [
                        SyntheticData(**item) for item in data["data"]
                    ]

                    # TODO: evolution

                    # TODO: review synthetic data
                    for data in synthetic_data:
                        golden = Golden(
                            input=data.input,
                            expectedOutput=data.expected_output,
                            context=context,
                        )
                        goldens.append(golden)

            self.synthetic_goldens.extend(goldens)

            return goldens

    # TODO
    def generate_goldens_from_docs(self, path: str):
        if self.multithreading:
            pass
        else:
            pass
        pass

    def save_as(self, file_type: str, directory: str):
        if file_type not in valid_file_types:
            raise ValueError(
                f"Invalid file type. Available file types to save as: {', '.join(type for type in valid_file_types)}"
            )

        if len(self.synthetic_goldens) == 0:
            raise ValueError(
                f"No synthetic goldens found. Please generate goldens before attempting to save data as {file_type}"
            )

        new_filename = (
            datetime.datetime.now().strftime("%Y%m%d_%H%M%S") + f".{file_type}"
        )

        if not os.path.exists(directory):
            os.makedirs(directory)

        full_file_path = os.path.join(directory, new_filename)

        if file_type == "json":
            with open(full_file_path, "w") as file:
                json_data = [
                    {
                        "input": golden.input,
                        "expected_output": golden.expected_output,
                        "context": golden.context,
                    }
                    for golden in self.synthetic_goldens
                ]
                json.dump(json_data, file, indent=4)

        elif file_type == "csv":
            with open(full_file_path, "w", newline="") as file:
                writer = csv.writer(file)
                writer.writerow(["input", "expected_output", "context"])
                for golden in self.synthetic_goldens:
                    context_str = "|".join(golden.context)
                    writer.writerow(
                        [golden.input, golden.expected_output, context_str]
                    )

        print(f"Synthetic goldens saved at {full_file_path}!")



class SynthesizerTemplate:
    @staticmethod
    def generate_synthetic_data(context, max_goldens_per_context):
        return f"""I want you act as a copywriter. Based on the given context, which is list of strings, please generate a list of JSON objects with the `input` and `expected_output` key.
The `input` can either be a question or a statement that can be addressed by the given context.
The `expected_output` is what an ideal output should look like for the corresponding generated input.
The `expected_output` should NEVER contradict the given context in any way.

**
IMPORTANT: Please make sure to only return in JSON format, with the 'data' key as a list of JSON objects.
You MUST TRY to generate {max_goldens_per_context} data points, unless there is too little context such that the `input` and `expected_output` is getting reptitive.

Example context: ["Einstein won the Nobel Prize for his discovery of penicillin.", "Einstein won the Nobel Prize in 1968."]
Example JSON:
{{
    "data": [
        {{
            "input": "What was Einstein known for?",
            "expected_output": "Einstein was known for his discovery of penicillin. He also won the Nobel Prize in 1968."
        }},
        {{
            "input": "Einstein was a smart guy huh",
            "expected_output": "He sure was! Einstein allegedly had an IQ of 200 and was the first to disocver penicillin."
        }}
    ]  
}}


You should NOT incorporate any prior knowledge you have and take each context at face value.
You should NOT be lazy and simply copy the context as the `expected_output`.
You MUST include at least one statement as the input.
Both `input` and `expected_output` are STRINGS.
You MUST TRY to generate {max_goldens_per_context} data points, unless there is too little context such that the `input` and `expected_output` is getting reptitive.
**

Context:
{context}

JSON:
"""


######################################################################################################
##### Templates taken from https://github.com/nlpxucan/WizardLM/blob/main/Evol_Instruct/depth.py #####
######################################################################################################

# TODO: Modify this

depth_base_instruction = "I want you act as a Prompt Rewriter.\r\n \
					Your objective is to rewrite a given prompt into a more complex version to make those famous AI systems (e.g., chatgpt and GPT4) a bit harder to handle.\r\n \
					But the rewritten prompt must be reasonable and must be understood and responded by humans.\r\n \
					Your rewriting cannot omit the non-text parts such as the table and code in #The Given Prompt#:. Also, please do not omit the input in #The Given Prompt#. \r\n \
					You SHOULD complicate the given prompt using the following method: \r\n\
					{} \r\n\
					You should try your best not to make the #Rewritten Prompt# become verbose, #Rewritten Prompt# can only add 10 to 20 words into #The Given Prompt#. \r\n\
					'#The Given Prompt#', '#Rewritten Prompt#', 'given prompt' and 'rewritten prompt' are not allowed to appear in #Rewritten Prompt#\r\n"

breadth_base_instruction = "I want you act as a Prompt Creator.\r\n\
Your goal is to draw inspiration from the #Given Prompt# to create a brand new prompt.\r\n\
This new prompt should belong to the same domain as the #Given Prompt# but be even more rare.\r\n\
The LENGTH and complexity of the #Created Prompt# should be similar to that of the #Given Prompt#.\r\n\
The #Created Prompt# must be reasonable and must be understood and responded by humans.\r\n\
'#Given Prompt#', '#Created Prompt#', 'given prompt' and 'created prompt' are not allowed to appear in #Created Prompt#\r\n"


class EvolutionTemplate:
    @staticmethod
    def generate_constraints_prompt(instruction):
        prompt = depth_base_instruction.format(
            "Please add one more constraints/requirements into #The Given Prompt#'"
        )
        prompt += "#The Given Prompt#: \r\n {} \r\n".format(instruction)
        prompt += "#Rewritten Prompt#:\r\n"
        return prompt

    @staticmethod
    def generate_deepen_prompt(instruction):
        prompt = depth_base_instruction.format(
            "If #The Given Prompt# contains inquiries about certain issues, the depth and breadth of the inquiry can be increased."
        )
        prompt += "#The Given Prompt#: \r\n {} \r\n".format(instruction)
        prompt += "#Rewritten Prompt#:\r\n"
        return prompt

    @staticmethod
    def generate_concretizing_prompt(instruction):
        prompt = depth_base_instruction.format(
            "Please replace general concepts with more specific concepts."
        )
        prompt += "#The Given Prompt#: \r\n {} \r\n".format(instruction)
        prompt += "#Rewritten Prompt#:\r\n"
        return prompt

    @staticmethod
    def generate_reasoning_prompt(instruction):
        prompt = depth_base_instruction.format(
            "If #The Given Prompt# can be solved with just a few simple thinking processes, you can rewrite it to explicitly request multiple-step reasoning."
        )
        prompt += "#The Given Prompt#: \r\n {} \r\n".format(instruction)
        prompt += "#Rewritten Prompt#:\r\n"
        return prompt

    @staticmethod
    def generate_breadth_prompt(instruction):
        prompt = breadth_base_instruction
        prompt += "#Given Prompt#: \r\n {} \r\n".format(instruction)
        prompt += "#Created Prompt#:\r\n"
        return prompt



from typing import Optional, List, Union
from pydantic import BaseModel, Field

from deepeval.utils import trimAndLoadJson
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.models import GPTModel, DeepEvalBaseLLM
from deepeval.metrics.answer_relevancy.template import AnswerRelevancyTemplate
from deepeval.progress_context import metrics_progress_context
from deepeval.telemetry import capture_metric_type


class AnswerRelvancyVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)


class AnswerRelevancyMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        if isinstance(model, DeepEvalBaseLLM):
            self.model = model
        else:
            self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.n = 5

    def measure(self, test_case: LLMTestCase) -> float:
        if (
            test_case.input is None
            or test_case.actual_output is None
            or test_case.retrieval_context is None
        ):
            raise ValueError(
                "Input, actual output, or retrieval context cannot be None"
            )
        with metrics_progress_context(self.__name__, self.evaluation_model):
            # generate statements
            self.statements: List[str] = self._generate_statements(
                test_case.actual_output
            )

            # generate verdicts based on statements, and retrieval context
            self.verdicts: List[AnswerRelvancyVerdict] = (
                self._generate_verdicts(
                    test_case.input, test_case.retrieval_context
                )
            )

            answer_relevancy_score = self._generate_score()

            self.reason = self._generate_reason(
                test_case.input, answer_relevancy_score
            )
            self.success = answer_relevancy_score >= self.threshold
            self.score = answer_relevancy_score
            capture_metric_type(self.__name__)
            return self.score

    def _generate_score(self):
        if len(self.verdicts) == 0:
            return 0

        relevant_count = 0
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() != "no":
                relevant_count += 1

        return relevant_count / len(self.verdicts)

    def _generate_reason(self, input: str, score: float) -> str:
        if self.include_reason is False:
            return None

        irrelevant_statements = []
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "no":
                irrelevant_statements.append(verdict.reason)

        prompt = AnswerRelevancyTemplate.generate_reason(
            irrelevant_statements=irrelevant_statements,
            input=input,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res

    def _generate_verdicts(
        self, input: str, retrieval_context=List[str]
    ) -> List[AnswerRelvancyVerdict]:
        prompt = AnswerRelevancyTemplate.generate_verdicts(
            input=input,
            actual_output=self.statements,
            retrieval_context="\n\n".join(retrieval_context),
        )

        res = self.model(prompt)
        data = trimAndLoadJson(res)
        verdicts = [AnswerRelvancyVerdict(**item) for item in data["verdicts"]]
        return verdicts

    def _generate_statements(
        self,
        actual_output: str,
    ) -> List[str]:
        prompt = AnswerRelevancyTemplate.generate_statements(
            actual_output=actual_output,
        )

        res = self.model(prompt)
        data = trimAndLoadJson(res)
        return data["statements"]

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Answer Relevancy"



_hyperparameters = None
_model = None
_user_prompt_template = None


def log_hyperparameters(model: str, prompt_template: str):
    def decorator(func):
        global _model, _user_prompt_template
        _model = model
        _user_prompt_template = prompt_template

        global _hyperparameters
        _hyperparameters = func()

        # Define the wrapper function that will be the actual decorator
        def wrapper(*args, **kwargs):
            # Optional: You can decide if you want to do something else here
            # every time the decorated function is called
            return func(*args, **kwargs)

        # Return the wrapper function to be used as the decorator
        return wrapper

    # Return the decorator itself
    return decorator


def get_hyperparameters():
    return _hyperparameters


def get_model():
    return _model


def get_user_prompt_template():
    return _user_prompt_template



class ContextualRelevancyTemplate:
    @staticmethod
    def generate_reason(input, irrelevant_sentences, score):
        return f"""Based on the given input, irrelevant sentences (list of JSON), and the contextual relevancy score (the closer to 1 the better), please generate a CONCISE reason for the score.
Irrelevant Sentences will contain JSONs with two keys: `sentence` and `node`. `sentence` is the actual sentence itself, and `node` is the node number from the `retrieval context` which it was drawn from. Specify that nodes are from retrieval context the first time you mention it.
In your reason, you should use data in the irrelevant sentences to support your point.

Contextual Relevancy Score:
{score}

Input:
{input}

Irrelevant Sentences:
{irrelevant_sentences}

Example:
The score is <contextual_relevancy_score> because <your_reason>.

** 
IMPORTANT:
If the score is 1, keep it short and say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).
**

Reason:
"""

    @staticmethod
    def generate_verdicts(text, context):
        return f"""Based on the input and context, please generate a list of JSON objects to indicate whether each given sentence in the context relevant to the provided input. The JSON will have 1 mandatory field: 'verdict', and 1 optional field: 'sentence'.
The 'verdict' key should STRICTLY be either 'yes' or 'no', and states whether the sentence is relevant to the text. 
Copy the sentence and supply the value to the 'sentence' key ONLY IF verdict is no.

**
IMPORTANT: Please make sure to only return in JSON format, with the 'verdicts' key as a list of JSON objects.
Example Context: "Einstein won the Nobel Prize for his discovery of the photoelectric effect. He won the Nobel Prize in 1968. There was a cat."
Example Input: "Einstein won the Nobel Prize in 1969 for his discovery of the photoelectric effect."

Example:
{{
    "verdicts": [
        {{
            "verdict": "yes"
        }},
        {{
            "verdict": "yes"
        }},
        {{
            "verdict": "no",
            "sentence": "There was a cat"
        }}
    ]  
}}
**

Input:
{text}

Context:
{context}

JSON:
"""



import os
from typing import List
import time
from dataclasses import dataclass

from deepeval.utils import drop_and_copy
from deepeval.telemetry import capture_evaluation_count
from deepeval.progress_context import progress_context
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.tracing import get_trace_stack
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.test_run import test_run_manager, APITestCase, MetricsMetadata
from deepeval.utils import get_is_running_deepeval


@dataclass
class TestResult:
    """Returned from run_test"""

    success: bool
    metrics: List[BaseMetric]
    input: str
    actual_output: str
    expected_output: str
    context: List[str]
    retrieval_context: List[str]


def create_test_result(
    test_case: LLMTestCase,
    success: bool,
    metrics: list[BaseMetric],
) -> TestResult:
    if isinstance(test_case, LLMTestCase):
        return TestResult(
            success=success,
            metrics=metrics,
            input=test_case.input,
            actual_output=test_case.actual_output,
            expected_output=test_case.expected_output,
            context=test_case.context,
            retrieval_context=test_case.retrieval_context,
        )
    else:
        raise ValueError("TestCase not supported yet.")


def execute_test(
    test_cases: List[LLMTestCase],
    metrics: List[BaseMetric],
    save_to_disk: bool = False,
) -> List[TestResult]:
    test_results: List[TestResult] = []
    test_run_manager.save_to_disk = save_to_disk
    for index, test_case in enumerate(test_cases):
        success = True
        api_test_case: APITestCase = APITestCase(
            name=os.getenv(PYTEST_RUN_TEST_NAME, f"test_case_{index}"),
            input=test_case.input,
            actualOutput=test_case.actual_output,
            expectedOutput=test_case.expected_output,
            success=success,
            metricsMetadata=[],
            runDuration=0,
            latency=test_case.latency,
            cost=test_case.cost,
            context=test_case.context,
            retrievalContext=test_case.retrieval_context,
            traceStack=get_trace_stack(),
            id=test_case.id,
        )

        test_start_time = time.perf_counter()
        for metric in metrics:
            metric.measure(test_case)
            metric_metadata = MetricsMetadata(
                metric=metric.__name__,
                score=metric.score,
                threshold=metric.threshold,
                reason=metric.reason,
                success=metric.is_successful(),
                evaluationModel=metric.evaluation_model,
            )
            api_test_case.metrics_metadata.append(metric_metadata)

            if metric_metadata.success is False:
                success = False

        test_end_time = time.perf_counter()
        run_duration = test_end_time - test_start_time
        api_test_case.run_duration = run_duration
        api_test_case.success = success

        test_run = test_run_manager.get_test_run()
        test_run.test_cases.append(api_test_case)
        test_run.dataset_alias = test_case.dataset_alias
        test_run_manager.save_test_run()

        test_result = create_test_result(
            test_case, success, drop_and_copy(metrics, ["model", "embeddings"])
        )
        test_results.append(test_result)

    return test_results


def run_test(
    test_case: LLMTestCase,
    metrics: List[BaseMetric],
) -> List[TestResult]:
    # TODO: refactor
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    # TODO: refactor
    if not isinstance(test_case, LLMTestCase):
        raise TypeError("'test_case' must be an instance of 'LLMTestCase'.")

    test_run_manager.reset()
    with progress_context("Executing run_test()..."):
        test_result = execute_test([test_case], metrics, False)[0]
        capture_evaluation_count()
        print_test_result(test_result)
        print("")
        print("-" * 70)
        return test_result


def assert_test(test_case: LLMTestCase, metrics: List[BaseMetric]):
    # TODO: refactor
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    # TODO: refactor
    if not isinstance(test_case, LLMTestCase):
        raise TypeError("'test_case' must be an instance of 'LLMTestCase'.")

    test_result = execute_test([test_case], metrics, get_is_running_deepeval())[
        0
    ]
    if not test_result.success:
        failed_metrics = [
            metric
            for metric in test_result.metrics
            if not metric.is_successful()
        ]
        failed_metrics_str = ", ".join(
            [
                f"{metric.__name__} (score: {metric.score}, threshold: {metric.threshold})"
                for metric in failed_metrics
            ]
        )
        raise AssertionError(f"Metrics {failed_metrics_str} failed.")


def evaluate(test_cases: List[LLMTestCase], metrics: List[BaseMetric]):
    # TODO: refactor
    for metric in metrics:
        if not isinstance(metric, BaseMetric):
            raise TypeError("Provided 'metric' must be of type 'BaseMetric'.")

    # TODO: refactor
    for test_case in test_cases:
        if not isinstance(test_case, LLMTestCase):
            raise TypeError(
                "Provided `test_cases` must be of type 'List[LLMTestCase]'."
            )

    test_run_manager.reset()
    with progress_context("Evaluating testcases..."):
        test_results = execute_test(test_cases, metrics, True)
        capture_evaluation_count()
        for test_result in test_results:
            print_test_result(test_result)
        print("")
        print("-" * 70)

        test_run_manager.wrap_up_test_run(display_table=False)
        return test_results


def print_test_result(test_result: TestResult):
    print("")
    print("=" * 70 + "\n")
    print("Metrics Summary\n")
    for metric in test_result.metrics:
        if not metric.is_successful():
            print(
                f"  - ❌ {metric.__name__} (score: {metric.score}, threshold: {metric.threshold}, evaluation model: {metric.evaluation_model}, reason: {metric.reason})"
            )
        else:
            print(
                f"  - ✅ {metric.__name__} (score: {metric.score}, threshold: {metric.threshold}, evaluation model: {metric.evaluation_model}, reason: {metric.reason})"
            )
        if metric.score_breakdown:
            for metric_name, score in metric.score_breakdown.items():
                print(f"      - {metric_name} (score: {score})")

    print("")
    print("For test case:\n")
    print(f"  - input: {test_result.input}")
    print(f"  - actual output: {test_result.actual_output}")
    print(f"  - expected output: {test_result.expected_output}")
    print(f"  - context: {test_result.context}")
    print(f"  - retrieval context: {test_result.retrieval_context}")



from deepeval.benchmarks.big_bench_hard.task import BigBenchHardTask


class BigBenchHardTemplate:

    @staticmethod
    def generate_output(input, task: BigBenchHardTask):
        if task == BigBenchHardTask.BOOLEAN_EXPRESSIONS:
            return f"""Given the boolean logic, output either 'True' or 'False' as the prediction. No explanations needed.

Boolean Logic:
{input}

Prediction:
"""
        else:
            # TODO: find or implement prompts to conform output
            pass



from typing import List, Optional, Union
from pydantic import BaseModel, Field

from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.progress_context import metrics_progress_context
from deepeval.telemetry import capture_metric_type
from deepeval.models import GPTModel, DeepEvalBaseLLM
from deepeval.utils import trimAndLoadJson
from deepeval.metrics.bias.template import BiasTemplate


# BiasMetric runs a similar algorithm to Dbias: https://arxiv.org/pdf/2208.05777.pdf
class BiasVerdict(BaseModel):
    verdict: str
    reason: str = Field(default=None)


class BiasMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
    ):
        self.threshold = threshold
        if isinstance(model, DeepEvalBaseLLM):
            self.model = model
        else:
            self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason

    def measure(self, test_case: LLMTestCase):
        if test_case.input is None or test_case.actual_output is None:
            raise ValueError("Input or actual output cannot be None")

        with metrics_progress_context(self.__name__, self.evaluation_model):
            self.opinions: List[str] = self._generate_opinions(
                test_case.actual_output
            )

            self.verdicts: List[BiasVerdict] = self._generate_verdicts()
            bias_score = self._generate_score()
            self.reason = self._generate_reason(bias_score)
            self.success = bias_score <= self.threshold
            self.score = bias_score

            capture_metric_type(self.__name__)
            return self.score

    def _generate_reason(self, score) -> str:
        if self.include_reason is False:
            return None

        biases = []
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "yes":
                biases.append(verdict.reason)

        prompt: dict = BiasTemplate.generate_reason(
            biases=biases,
            score=format(score, ".2f"),
        )

        res = self.model(prompt)
        return res

    def _generate_score(self) -> float:
        total = len(self.verdicts)
        if total == 0:
            return 0

        bias_count = 0
        for verdict in self.verdicts:
            if verdict.verdict.strip().lower() == "yes":
                bias_count += 1

        return bias_count / total

    def _generate_verdicts(self) -> List[BiasVerdict]:
        verdicts: List[BiasVerdict] = []

        prompt = BiasTemplate.generate_verdicts(opinions=self.opinions)
        res = self.model(prompt)
        data = trimAndLoadJson(res)

        verdicts = [BiasVerdict(**item) for item in data["verdicts"]]

        return verdicts

    def _generate_opinions(self, actual_output: str) -> List[str]:
        prompt = BiasTemplate.generate_opinions(actual_output=actual_output)
        res = self.model(prompt)
        data = trimAndLoadJson(res)

        return data["opinions"]

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Bias"



from typing import List, Optional, Union
from enum import Enum
from pydantic import BaseModel, Field
from concurrent.futures import ThreadPoolExecutor

from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.models import GPTModel, DeepEvalBaseLLM
from deepeval.utils import trimAndLoadJson
from deepeval.metrics.summarization.template import SummarizationTemplate
from deepeval.metrics.faithfulness.template import FaithfulnessTemplate
from deepeval.progress_context import metrics_progress_context
from deepeval.telemetry import capture_metric_type


class SummarizationAlignmentVerdict(BaseModel):
    # yes, no, or idk
    verdict: str
    reason: str = Field(default=None)


class SummarizationCoverageVerdict(BaseModel):
    summary_verdict: str
    original_verdict: str
    question: str = Field(default=None)


class ScoreType(Enum):
    ALIGNMENT = "Alignment"
    COVERAGE = "Coverage"


class SummarizationMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        n: int = 5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        assessment_questions: Optional[List[str]] = None,
        include_reason: bool = True,
        multithreading=True,
    ):
        self.threshold = threshold
        if isinstance(model, DeepEvalBaseLLM):
            self.model = model
        else:
            self.model = GPTModel(model=model)
        self.evaluation_model = self.model.get_model_name()

        if assessment_questions is not None and len(assessment_questions) == 0:
            self.assessment_questions = None
        else:
            self.assessment_questions = assessment_questions

        self.multithreading = multithreading
        self.include_reason = include_reason
        self.n = n

    def measure(self, test_case: LLMTestCase):
        if test_case.input is None or test_case.actual_output is None:
            raise ValueError("Input or actual output cannot be None")

        with metrics_progress_context(self.__name__, self.evaluation_model):
            if test_case.input is None or test_case.actual_output is None:
                raise ValueError("Input and actual output cannot be None")

            if self.multithreading:
                # Use multithreading to generate truths and claims in parallel
                with ThreadPoolExecutor() as executor:
                    future_truths = executor.submit(
                        # Claims made in the original text === truths
                        self._generate_claims,
                        test_case.input,
                    )
                    future_claims = executor.submit(
                        self._generate_claims, test_case.actual_output
                    )
                    future_coverage_verdicts = executor.submit(
                        self._generate_coverage_verdicts, test_case
                    )

                    self.truths: List[str] = future_truths.result()
                    self.claims: List[str] = future_claims.result()
                    self.coverage_verdicts: List[
                        SummarizationCoverageVerdict
                    ] = future_coverage_verdicts.result()
            else:
                # Sequential execution
                self.truths: List[str] = self._generate_claims(test_case.input)
                self.claims: List[str] = self._generate_claims(
                    test_case.actual_output
                )
                self.coverage_verdicts: List[SummarizationCoverageVerdict] = (
                    self._generate_coverage_verdicts(test_case)
                )

            self.alignment_verdicts: List[SummarizationAlignmentVerdict] = (
                self._generate_alignment_verdicts()
            )
            alignment_score = self._generate_score(ScoreType.ALIGNMENT)
            coverage_score = self._generate_score(ScoreType.COVERAGE)

            self.score_breakdown = {
                ScoreType.ALIGNMENT.value: alignment_score,
                ScoreType.COVERAGE.value: coverage_score,
            }
            summarization_score = min(alignment_score, coverage_score)
            self.reason = self._generate_reason(summarization_score)
            self.success = summarization_score >= self.threshold
            self.score = summarization_score
            capture_metric_type(self.__name__)
            return self.score

    def _generate_reason(self, score: float) -> str:
        if self.include_reason is False:
            return None

        contradictions = []
        redundancies = []
        for verdict in self.alignment_verdicts:
            if verdict.verdict.strip().lower() == "no":
                contradictions.append(verdict.reason)
            elif verdict.verdict.strip().lower() == "idk":
                redundancies.append(verdict.reason)

        questions = []
        if self.coverage_verdicts:
            for verdict in self.coverage_verdicts:
                if (
                    verdict.original_verdict.strip().lower() == "yes"
                    and verdict.summary_verdict.strip().lower() == "no"
                ):
                    questions.append(verdict.question)

        prompt: dict = SummarizationTemplate.generate_reason(
            contradictions=contradictions,
            redundancies=redundancies,
            questions=questions,
            score=format(score, ".2f"),
        )

        if len(questions) > 0:
            prompt += f"""Questions the original text can answer but not the summary:
{questions}

"""
        prompt += """Reason:"""

        res = self.model(prompt)
        return res

    def _generate_score(self, score_type: ScoreType) -> float:
        if score_type == ScoreType.ALIGNMENT:
            total = len(self.alignment_verdicts)
            if total == 0:
                return 0
            faithfulness_count = 0
            for verdict in self.alignment_verdicts:
                # Different from the faithfulness score, this
                # penalizes 'idk' (full of fluff) summaries
                if verdict.verdict.strip().lower() == "yes":
                    faithfulness_count += 1

            return faithfulness_count / total

        else:
            if self.assessment_questions is None:
                return 1
            total = 0
            coverage_count = 0
            for verdict in self.coverage_verdicts:
                if verdict.original_verdict.strip().lower() == "yes":
                    total += 1
                    if verdict.summary_verdict.strip().lower() == "yes":
                        coverage_count += 1

            if total == 0:
                return 0

            return coverage_count / total

    def _generate_answers(self, text: str) -> List[str]:
        prompt = SummarizationTemplate.generate_answers(
            questions=self.assessment_questions, text=text
        )
        res = self.model(prompt)
        data = trimAndLoadJson(res)
        return data["answers"]

    def _generate_assessment_questions(self, text: str):
        prompt = SummarizationTemplate.generate_questions(text=text, n=self.n)
        res = self.model(prompt)
        data = trimAndLoadJson(res)
        return data["questions"]

    def _generate_coverage_verdicts(
        self, test_case: LLMTestCase
    ) -> List[SummarizationCoverageVerdict]:
        if self.assessment_questions is None:
            self.assessment_questions = self._generate_assessment_questions(
                test_case.input
            )

        if self.multithreading:
            with ThreadPoolExecutor() as executor:
                future_original_answers: List[str] = executor.submit(
                    self._generate_answers, test_case.input
                )
                future_summary_answers: List[str] = executor.submit(
                    self._generate_answers, test_case.actual_output
                )
                original_answers = future_original_answers.result()
                summary_answers = future_summary_answers.result()

        else:
            original_answers = self._generate_answers(test_case.input)
            summary_answers = self._generate_answers(test_case.actual_output)

        if len(original_answers) != len(summary_answers):
            raise ValueError("Number of verdicts generated does not equal.")

        coverage_veridcts: List[SummarizationCoverageVerdict] = []
        for i in range(len(original_answers)):
            coverage_veridcts.append(
                SummarizationCoverageVerdict(
                    summary_verdict=summary_answers[i],
                    original_verdict=original_answers[i],
                    question=self.assessment_questions[i],
                )
            )

        return coverage_veridcts

    def _generate_alignment_verdicts(
        self,
    ) -> List[SummarizationAlignmentVerdict]:
        verdicts: List[SummarizationAlignmentVerdict] = []
        prompt = SummarizationTemplate.generate_alignment_verdicts(
            summary_claims=self.claims, orignal_text="\n\n".join(self.truths)
        )
        res = self.model(prompt)
        data = trimAndLoadJson(res)
        verdicts = [
            SummarizationAlignmentVerdict(**item) for item in data["verdicts"]
        ]

        return verdicts

    def _generate_claims(self, text: str) -> List[str]:
        # Borrow faithfulness template
        prompt = FaithfulnessTemplate.generate_claims(text=text)
        res = self.model(prompt)
        data = trimAndLoadJson(res)

        return data["claims"]

    def is_successful(self) -> bool:
        self.success = self.score >= self.threshold
        return self.success

    @property
    def __name__(self):
        return "Summarization"
